{-# LANGUAGE GeneralizedNewtypeDeriving #-}

-- Monoid (IO Monoid) and MonadPlus IO

module IOPlus () where

import Data.Semigroup
import Control.Applicative
import Control.Monad
import System.IO.Error (ioeGetErrorString, catchIOError)

handleOnly name result = \e -> if ioeGetErrorString e == name then result else ioError e

-- IO.mplus evaluates only until first non-zero
instance Semigroup a => Semigroup (IO a) where { (<>) = liftM2 (<>) }
instance (Semigroup a, Monoid a) => Monoid (IO a) where { mempty = return mempty; mappend = (<>) }
instance MonadPlus IO where { mzero = fail "IOPlus.mzero"; mplus u v = catchIOError u (handleOnly "IOPlus.mzero" v) }
instance Alternative IO where { empty = mzero; u <|> v = u `mplus` v }

{-
-- IOPlus.mplus evaluates only until first non-zero
newtype IOPlus a = IOPlus { getIOPlus :: IO a } deriving (Functor, Applicative, Monad)
instance Semigroup a => Semigroup (IOPlus a) where { u <> v = IOPlus $ getIOPlus u >>= \u' -> getIOPlus v >>= \v' -> return (u' <> v') }
instance (Semigroup a, Monoid a) => Monoid (IOPlus a) where { mempty = IOPlus $ return mempty; mappend u v = u <> v }
instance MonadPlus IOPlus where { mzero = IOPlus (fail "IOPlus.mzero"); mplus u v = IOPlus $ catchIOError (getIOPlus u) (handleOnly "IOPlus.mzero" $ getIOPlus v) }
instance Alternative IOPlus where { empty = mzero; u <|> v = u `mplus` v }
-}

-- IOFirst.mplus evaluates the whole sequence, returns first non-zero
newtype IOFirst a = IOFirst { getIOFirst :: IO a } deriving (Functor, Applicative, Monad)
instance Semigroup a => Semigroup (IOFirst a) where { u <> v = IOFirst $ getIOFirst u >>= \u' -> getIOFirst v >>= \v' -> return (u' <> v') }
instance (Semigroup a, Monoid a) => Monoid (IOFirst a) where { mempty = IOFirst $ return mempty; mappend u v = u <> v }
instance MonadPlus IOFirst where { mzero = IOFirst $ fail "IOFirst.mzero"; mplus u v = IOFirst $ catchIOError (getIOFirst u >>= \u' -> catchIOError (getIOFirst v >> return u') (handleOnly "IOFirst.mzero" $ return u')) (handleOnly "IOFirst.mzero" $ getIOFirst v) }
instance Alternative IOFirst where { empty = mzero; u <|> v = u `mplus` v }

-- IOLast.mplus evaluates the whole sequence, returns last non-zero
newtype IOLast a = IOLast { getIOLast :: IO a } deriving (Functor, Applicative, Monad)
instance Semigroup a => Semigroup (IOLast a) where { u <> v = IOLast $ getIOLast u >>= \u' -> getIOLast v >>= \v' -> return (u' <> v') }
instance (Semigroup a, Monoid a) => Monoid (IOLast a) where { mempty = IOLast $ return mempty; mappend u v = u <> v }
instance MonadPlus IOLast where { mzero = IOLast (fail "IOLast.mzero"); mplus u v = IOLast $ catchIOError (getIOLast u >>= \u' -> catchIOError (getIOLast v) (handleOnly "IOLast.mzero" $ return u')) (handleOnly "IOLast.mzero" $ getIOLast v) }
instance Alternative IOLast where { empty = mzero; u <|> v = u `mplus` v }
